/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/
#include "un_hl_ve.h"

int UN_HL_VE_CL_GROW = 0;
int UN_HL_VE_CL_MAX_SZ = 20;

int stats_un_hl_ve_cl = 0;


int get_stats_un_hl_ve() {
  return stats_un_hl_ve_cl;
}

/* ========================================================================== */
/*                           Prop_tr_un_fl functions                          */
/* ========================================================================== */
prop_tr_un_fl* build_not_prop_tr_un_fl(prop_tr_un_fl *cloned_ptuf) {
  prop_tr_un_fl *new_ptuf = new prop_tr_un_fl();

  if (cloned_ptuf->op == VAR) {
    new_ptuf->op = cloned_ptuf->op;
    new_ptuf->lit = aiger_not(cloned_ptuf->lit);
    new_ptuf->lit_ref = cloned_ptuf->lit_ref;

  } else {
    if (cloned_ptuf->op == AND)
      new_ptuf->op = OR;
    else {
      assert(cloned_ptuf->op == OR);
      new_ptuf->op = AND;      
    }
    new_ptuf->lp = new list_ptuf();
    list_ptuf_e *it = cloned_ptuf->lp->head;
    while(it != NULL) {

      prop_tr_un_fl *ptuf_tmp = build_not_prop_tr_un_fl(it->ptuf);
      insert_ordered(new_ptuf->lp, ptuf_tmp);

      it = it->next;
    }
  }

  return new_ptuf;
}

prop_tr_un_fl* clone_prop_tr_un_fl(prop_tr_un_fl *cloned_ptuf) {
  prop_tr_un_fl *new_ptuf = new prop_tr_un_fl();
  new_ptuf->op = cloned_ptuf->op;
  if (cloned_ptuf->op == VAR) {
    new_ptuf->lit = cloned_ptuf->lit;
    new_ptuf->lit_ref = cloned_ptuf->lit_ref;

  } else {

    new_ptuf->lp = new list_ptuf();
    list_ptuf_e *it = cloned_ptuf->lp->head;
    while(it != NULL) {

      prop_tr_un_fl *ptuf_tmp = clone_prop_tr_un_fl(it->ptuf);
      insert_ordered(new_ptuf->lp, ptuf_tmp);

      it = it->next;
    }
  }

  return new_ptuf;
}

void elim_prop_tr_un_fl(prop_tr_un_fl *ptuf, prop_tr_un_fl *ptuf_rep_with, int var2rep) {
  list_ptuf_e *it = ptuf->lp->head;
  while(it != NULL) {

    if (it->ptuf->op == VAR) {
      if (aiger_lit2var(it->ptuf->lit) == var2rep) {
	remove(ptuf->lp, it);
	prop_tr_un_fl *ptuf_tmp;
	if (aiger_sign(it->ptuf->lit)) {
	  // Build not_ptuf_rep_with & Replace ptuf by not_ptuf_rep_with
	  ptuf_tmp = build_not_prop_tr_un_fl(ptuf_rep_with);
	  if (ptuf_tmp->op == ptuf->op) {
	    list_ptuf_e *it_ = ptuf_tmp->lp->head;
	    while(it_ != NULL) {
	      insert_ordered(ptuf->lp, it_->ptuf);
	      it_ = it_->next;
	    }
	  } else {
	    insert_ordered(ptuf->lp, ptuf_tmp);
	  }
	} else {
	  // Replace ptuf by ptuf_rep_with 
	  ptuf_tmp = clone_prop_tr_un_fl(ptuf_rep_with);
	  if (ptuf_tmp->op == ptuf->op) {
	    list_ptuf_e *it_ = ptuf_tmp->lp->head;
	    while(it_ != NULL) {
	      insert_ordered(ptuf->lp, it_->ptuf);
	      it_ = it_->next;
	    }
	  } else {
	    insert_ordered(ptuf->lp, ptuf_tmp);
	  }
	}
      }
    } else {
      elim_prop_tr_un_fl(it->ptuf, ptuf_rep_with, var2rep);
    }

    it = it->next;
  }
}



/* ========================================================================== */
/*                          High level VE functions                           */
/* ========================================================================== */
std::pair<std::list<std::list<int>>, std::list<std::list<int>>> un_compute_cls_gen_exact(prop_tr_un_fl *ptuf) {
  std::list<int> li;
  std::list<std::list<int>> lli_f;
  std::list<std::list<int>> lli_s;
  
  if (ptuf->op == VAR) {
    li.push_back(ptuf->lit);
    lli_f.push_back(li);
    lli_s.push_back(li);

  } else {
    std::list<std::pair<std::list<std::list<int>>, std::list<std::list<int>>>> lli_pl;
    
    list_ptuf_e *it = ptuf->lp->head;
    
    while(it != NULL) {
      std::pair<std::list<std::list<int>>, std::list<std::list<int>>> lli_p = un_compute_cls_gen_exact(it->ptuf);
      lli_pl.push_back(lli_p);
      
      it = it->next;
    }

    std::list<std::pair<std::list<std::list<int>>, std::list<std::list<int>>>>::iterator it_llip;
    std::list<std::list<int>>::iterator it_lli;
    std::list<std::list<int>>::iterator it_lli2;
    std::list<int>::iterator it_li;

    if (ptuf->op == AND) {
      
      for( it_llip = lli_pl.begin(); it_llip != lli_pl.end(); it_llip++) {
	//std::list<std::list<int>> lli_op_f = it_llip->first;
	for (it_lli = it_llip->first.begin(); it_lli != it_llip->first.end(); it_lli++) {
	  lli_f.push_back(*it_lli);
	}
      }
      
      std::list<std::list<int>> lli_op1_s = lli_pl.front().second;
      lli_pl.pop_front();
      while (!lli_pl.empty()) {
	std::list<std::list<int>> lli_op2_s = lli_pl.front().second;
	lli_pl.pop_front();
	for (it_lli = lli_op1_s.begin(); it_lli != lli_op1_s.end(); it_lli++) {
	  for (it_lli2 = lli_op2_s.begin(); it_lli2 != lli_op2_s.end(); it_lli2++) {
	    li.clear();
	    
	    for (it_li = it_lli->begin(); it_li != it_lli->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    for (it_li = it_lli2->begin(); it_li != it_lli2->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    lli_s.push_back(li);
	  }
	}
	lli_op1_s = lli_s;
	lli_s.clear();
      }
      lli_s = lli_op1_s;
      
    } else {

      for( it_llip = lli_pl.begin(); it_llip != lli_pl.end(); it_llip++) {
	//std::list<std::list<int>> lli_op_s = it_llip->second;
	for (it_lli = it_llip->second.begin(); it_lli != it_llip->second.end(); it_lli++) {
	  lli_s.push_back(*it_lli);
	}

      }
      
      std::list<std::list<int>> lli_op1_f = lli_pl.front().first;
      lli_pl.pop_front();
      while (!lli_pl.empty()) {
	std::list<std::list<int>> lli_op2_f = lli_pl.front().first;
	lli_pl.pop_front();
	for (it_lli = lli_op1_f.begin(); it_lli != lli_op1_f.end(); it_lli++) {
	  for (it_lli2 = lli_op2_f.begin(); it_lli2 != lli_op2_f.end(); it_lli2++) {
	    li.clear();
	    
	    for (it_li = it_lli->begin(); it_li != it_lli->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    for (it_li = it_lli2->begin(); it_li != it_lli2->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    lli_f.push_back(li);
	  }
	}
	lli_op1_f = lli_f;
	lli_f.clear();
      }
      lli_f = lli_op1_f;
      
    }
  }
  return std::pair<std::list<std::list<int>>, std::list<std::list<int>>>(lli_f, lli_s);
}

std::pair<std::pair<int, int>, std::pair<int, int>> un_compute_nb_cls_gen_exact(prop_tr_un_fl *ptuf) {
  std::pair<std::pair<int, int>, std::pair<int, int>> ret = std::pair<std::pair<int, int>, std::pair<int, int>>(std::pair<int, int>(0, 0), std::pair<int, int>(0,0));
  std::set<int> set_cl;
  std::pair<std::list<std::list<int>>, std::list<std::list<int>>> list_of_cls = un_compute_cls_gen_exact(ptuf);
  std::list<std::list<int>> cls = list_of_cls.first;
  int nb_of_cls = 0;
  for (std::list<std::list<int>>::iterator it = cls.begin(); it != cls.end(); it++) {
    bool trivial_cl = 0;
    for (std::list<int>::iterator it_lit = (*it).begin(); it_lit != (*it).end(); it_lit++) {
      set_cl.insert(*it_lit);
    }
    ret.second.second = ret.second.second > set_cl.size() ? ret.second.second : set_cl.size();
    for (std::list<int>::iterator it_lit = (*it).begin(); !trivial_cl && it_lit != --(*it).end();) {
      int current = *it_lit;
      it_lit++;
      if (aiger_lit2var(current) == aiger_lit2var(*it_lit))
	trivial_cl = 1;
    }
    if (!trivial_cl)
      nb_of_cls++;
    set_cl.clear();
  }
  ret.first.second = nb_of_cls;

  cls = list_of_cls.second;
  nb_of_cls = 0;
  for (std::list<std::list<int>>::iterator it = cls.begin(); it != cls.end(); it++) {
    bool trivial_cl = 0;
    for (std::list<int>::iterator it_lit = (*it).begin(); it_lit != (*it).end(); it_lit++) {
      set_cl.insert(*it_lit);
    }
    ret.second.first = ret.second.first > set_cl.size() ? ret.second.first : set_cl.size();
    for (std::list<int>::iterator it_lit = (*it).begin(); !trivial_cl && it_lit != --(*it).end();) {
      int current = *it_lit;
      it_lit++;
      if (aiger_lit2var(current) == aiger_lit2var(*it_lit))
	trivial_cl = 1;
    }
    if (!trivial_cl)
      nb_of_cls++;
    set_cl.clear();
  }
  ret.first.first = nb_of_cls;
  return ret;
}



std::pair<std::list<std::list<int>>, std::list<std::list<int>>> un_compute_cls_gen_ve_exact
  (prop_tr_un_fl *ptuf, int var_to_elim,
   std::pair<std::list<std::list<int>>, std::list<std::list<int>>> cl_gen_var) {
    std::list<int> li;
    std::list<std::list<int>> lli_f;
    std::list<std::list<int>> lli_s;
    
    if (ptuf->op == VAR) {
      if (aiger_lit2var(ptuf->lit) == var_to_elim)
	if (aiger_sign(ptuf->lit))
	  return std::pair<std::list<std::list<int>>, std::list<std::list<int>>>(cl_gen_var.second, cl_gen_var.first);
      else
	return cl_gen_var;
      else {
	li.push_back(ptuf->lit);
	lli_f.push_back(li);
	lli_s.push_back(li);
      }
    } else {
      std::list<std::pair<std::list<std::list<int>>, std::list<std::list<int>>>> lli_pl;
      
      list_ptuf_e *it = ptuf->lp->head;
    
      while(it != NULL) {
	std::pair<std::list<std::list<int>>, std::list<std::list<int>>> lli_p =
	  un_compute_cls_gen_ve_exact(it->ptuf, var_to_elim, cl_gen_var);
	lli_pl.push_back(lli_p);
	
	it = it->next;
      }
    
      std::list<std::pair<std::list<std::list<int>>, std::list<std::list<int>>>>::iterator it_llip;
      std::list<std::list<int>>::iterator it_lli;
      std::list<std::list<int>>::iterator it_lli2;
      std::list<int>::iterator it_li;

      if (ptuf->op == AND) {
      
	for( it_llip = lli_pl.begin(); it_llip != lli_pl.end(); it_llip++) {
	  //std::list<std::list<int>> lli_op_f = it_llip->first;
	  for (it_lli = it_llip->first.begin(); it_lli != it_llip->first.end(); it_lli++) {
	    lli_f.push_back(*it_lli);
	  }
	}
      
	std::list<std::list<int>> lli_op1_s = lli_pl.front().second;
	lli_pl.pop_front();
	while (!lli_pl.empty()) {
	  std::list<std::list<int>> lli_op2_s = lli_pl.front().second;
	  lli_pl.pop_front();
	  for (it_lli = lli_op1_s.begin(); it_lli != lli_op1_s.end(); it_lli++) {
	    for (it_lli2 = lli_op2_s.begin(); it_lli2 != lli_op2_s.end(); it_lli2++) {
	      li.clear();
	    
	      for (it_li = it_lli->begin(); it_li != it_lli->end(); it_li++) {
		li.push_back(*it_li);
	      }
	      for (it_li = it_lli2->begin(); it_li != it_lli2->end(); it_li++) {
		li.push_back(*it_li);
	      }
	      lli_s.push_back(li);
	    }
	  }
	  lli_op1_s = lli_s;
	  lli_s.clear();
	}
	lli_s = lli_op1_s;
      
      } else {

	for( it_llip = lli_pl.begin(); it_llip != lli_pl.end(); it_llip++) {
	  //std::list<std::list<int>> lli_op_s = it_llip->second;
	  for (it_lli = it_llip->second.begin(); it_lli != it_llip->second.end(); it_lli++) {
	    lli_s.push_back(*it_lli);
	  }

	}
      
	std::list<std::list<int>> lli_op1_f = lli_pl.front().first;
	lli_pl.pop_front();
	while (!lli_pl.empty()) {
	  std::list<std::list<int>> lli_op2_f = lli_pl.front().first;
	  lli_pl.pop_front();
	  for (it_lli = lli_op1_f.begin(); it_lli != lli_op1_f.end(); it_lli++) {
	    for (it_lli2 = lli_op2_f.begin(); it_lli2 != lli_op2_f.end(); it_lli2++) {
	      li.clear();
	    
	      for (it_li = it_lli->begin(); it_li != it_lli->end(); it_li++) {
		li.push_back(*it_li);
	      }
	      for (it_li = it_lli2->begin(); it_li != it_lli2->end(); it_li++) {
		li.push_back(*it_li);
	      }
	      lli_f.push_back(li);
	    }
	  }
	  lli_op1_f = lli_f;
	  lli_f.clear();
	}
	lli_f = lli_op1_f;
      
      }
    }
  return std::pair<std::list<std::list<int>>, std::list<std::list<int>>>(lli_f, lli_s);
}

std::pair<std::pair<int, int>, std::pair<int, int>> un_compute_nb_cls_gen_ve_exact(prop_tr_un_fl *ptuf, int var_to_elim, prop_tr_un_fl *ptuf_repBy) {
  std::pair<std::pair<int, int>, std::pair<int, int>> ret = std::pair<std::pair<int, int>, std::pair<int, int>>(std::pair<int, int>(0, 0), std::pair<int, int>(0,0));
  std::set<int> set_cl;
  std::pair<std::list<std::list<int>>, std::list<std::list<int>>> cl_gen_var = un_compute_cls_gen_exact(ptuf_repBy);
  std::pair<std::list<std::list<int>>, std::list<std::list<int>>> list_of_cls = un_compute_cls_gen_ve_exact(ptuf, var_to_elim, cl_gen_var);
  std::list<std::list<int>> cls = list_of_cls.first;
  int nb_of_cls = 0;
  for (std::list<std::list<int>>::iterator it = cls.begin(); it != cls.end(); it++) {
    bool trivial_cl = 0;
    for (std::list<int>::iterator it_lit = (*it).begin(); it_lit != (*it).end(); it_lit++) {
      set_cl.insert(*it_lit);
    }
    ret.second.second = ret.second.second > set_cl.size() ? ret.second.second : set_cl.size();
    for (std::list<int>::iterator it_lit = (*it).begin(); !trivial_cl && it_lit != --(*it).end();) {
      int current = *it_lit;
      it_lit++;
      if (aiger_lit2var(current) == aiger_lit2var(*it_lit))
	trivial_cl = 1;
    }
    if (!trivial_cl)
      nb_of_cls++;
    set_cl.clear();
  }
  ret.first.second = nb_of_cls;

  cls = list_of_cls.second;
  nb_of_cls = 0;
  for (std::list<std::list<int>>::iterator it = cls.begin(); it != cls.end(); it++) {
    bool trivial_cl = 0;
    for (std::list<int>::iterator it_lit = (*it).begin(); it_lit != (*it).end(); it_lit++) {
      set_cl.insert(*it_lit);
    }
    ret.second.first = ret.second.first > set_cl.size() ? ret.second.first : set_cl.size();
    for (std::list<int>::iterator it_lit = (*it).begin(); !trivial_cl && it_lit != --(*it).end();) {
      int current = *it_lit;
      it_lit++;
      if (aiger_lit2var(current) == aiger_lit2var(*it_lit))
	trivial_cl = 1;
    }
    if (!trivial_cl)
      nb_of_cls++;
    set_cl.clear();
  }
  ret.first.first = nb_of_cls;
  return ret;
}



/********************************************************************************/

void update_refby_un_hl_ve(prop_tr_un_fl *ptuf, ands_def *ad_to_del, ands_def *ad_to_add) {
  assert(ad_to_del != ad_to_add);
  if (ptuf->op == VAR) {
    if (ptuf->lit_ref != NULL) {
      ptuf->lit_ref->refBy.erase (ad_to_del);
      ptuf->lit_ref->refBy.insert(ad_to_add);
    }
  } else {
    list_ptuf_e *it = ptuf->lp->head;
    while(it != NULL) {
      update_refby_un_hl_ve(it->ptuf, ad_to_del, ad_to_add);
      it = it->next;
    }
  }

}

/* ========================================================================== */
/*                                      VE                                    */
/* ========================================================================== */
void set_ad_apply_ve(int time) {
  idef* tmp_def;
  dll_elem<idef> *tmp_e;
  tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    assert(tmp_def->rhs.size() == 1);
      
    if (tmp_def->rhs.front() > 1) {
      assert(tmp_def->lhs_ts[time] != -1);

      idef *tmp_id = mapdef[aiger_lit2var(tmp_def->rhs.front())];
      int next = tmp_id->lhs_ts[time];

      if (next > 1) {
	std::map<unsigned, ands_def*>::iterator it_lit_ref = map_ands.find(aiger_lit2var(next));

	if (it_lit_ref != map_ands.end()) {
	  if (it_lit_ref->second->cst != -1) {
	    if (it_lit_ref->second->cst_ref != NULL) {
	      assert(it_lit_ref->second->cst_ref->cst == -1);
	      it_lit_ref->second->cst_ref->apply_ve = 0;
	    }
	  } else {
	    it_lit_ref->second->apply_ve = 0;
	  }
	}

	// else { Def <=> Input or Cst (=> Do nothing) }

      }
    }

    tmp_e = tmp_e->next;
  }
}

void reset_ad_apply_ve(int time) {
  ands_def* tmp_def;
  dll_elem<ands_def> *tmp_e;

  // Ands def
  tmp_e = list_ands[time]->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    tmp_def->apply_ve = 1;

    tmp_e = tmp_e->next;
  }
}


// std::set<int> test_elim_set;
// void test_elim(prop_tr_un_fl *ptuf) {
//   if (ptuf->op == VAR) {
//     if (test_elim_set.find(aiger_strip(ptuf->lit)) != test_elim_set.end())
//       printf("ERROR : %d\n", ptuf->lit);
//     assert(test_elim_set.find(aiger_strip(ptuf->lit)) == test_elim_set.end());
//   } else {
//     list_ptuf_e *it = ptuf->lp->head;
//     while(it != NULL) {
//       test_elim(it->ptuf);
//       it = it->next;
//     }
//   }
// }


void un_hl_ve_ClausesBased(int time) {
  ands_def *tmp_def, *def_refBy;
  dll_elem<ands_def> *tmp_e;
  std::set<ands_def*>::iterator it_ref;


  // Ands def
  tmp_e = list_ands[time]->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    /* Temporary heuristic to limit the time spent here */
    int temp_heuristic, max;
    max = 0;
    for (it_ref = tmp_def->refBy.begin(); it_ref != tmp_def->refBy.end(); ++it_ref)
      max = max > (*it_ref)->nb_op ? max : (*it_ref)->nb_op;
    temp_heuristic = tmp_def->nb_op + max;
    /************************************************/

    //test_elim(tmp_def->rhs_tr_fl);

    if (tmp_def->apply_ve && tmp_def->refBy.size() //) {
	&& temp_heuristic < 20) {

      std::pair<std::pair<int, int>, std::pair<int, int>> cl_nb_and_sz_tmp_def;
      cl_nb_and_sz_tmp_def = un_compute_nb_cls_gen_exact(tmp_def->rhs_tr_fl);

      std::pair<int, int> nb_of_cl_tmp_def = cl_nb_and_sz_tmp_def.first;
      std::pair<int, int> cl_sz_tmp_def = cl_nb_and_sz_tmp_def.second;

      int current_nb_of_cls = nb_of_cl_tmp_def.first + nb_of_cl_tmp_def.second;

      for (it_ref = tmp_def->refBy.begin(); it_ref != tmp_def->refBy.end(); ++it_ref) {
	def_refBy = *it_ref;
	std::pair<std::pair<int, int>, std::pair<int, int>> cl_nb_and_sz_ref;

	cl_nb_and_sz_ref = un_compute_nb_cls_gen_exact(def_refBy->rhs_tr_fl);

	std::pair<int, int> nb_of_cl_ref = cl_nb_and_sz_ref.first;
	current_nb_of_cls += nb_of_cl_ref.first + nb_of_cl_ref.second;
      }

      int future_nb_of_cls = 0;
      bool to_elim = 1;

      for (it_ref = tmp_def->refBy.begin(); to_elim && it_ref != tmp_def->refBy.end(); ++it_ref) {
      	def_refBy = *it_ref;
      	std::pair<std::pair<int, int>, std::pair<int, int>> cl_nb_and_sz_refBy;

	cl_nb_and_sz_refBy = un_compute_nb_cls_gen_ve_exact(def_refBy->rhs_tr_fl,
							    aiger_lit2var(tmp_def->lhs), tmp_def->rhs_tr_fl);

      	std::pair<int, int> nb_of_cl_def_refBy = cl_nb_and_sz_refBy.first;
      	std::pair<int, int> cl_sz_refBy = cl_nb_and_sz_refBy.second;

      	int max_sz_cl = cl_sz_refBy.first > cl_sz_refBy.second ? cl_sz_refBy.first+1 : cl_sz_refBy.second+1;

      	future_nb_of_cls += nb_of_cl_def_refBy.first + nb_of_cl_def_refBy.second;
      	if (future_nb_of_cls > current_nb_of_cls + UN_HL_VE_CL_GROW ||
      	    max_sz_cl > UN_HL_VE_CL_MAX_SZ)
      	  to_elim = 0;
      }

      if (to_elim) { 
	//test_elim_set.insert(tmp_def->lhs);
	//printf("%d, ELIM : \n", tmp_def->lhs);
	for (it_ref = tmp_def->refBy.begin(); it_ref != tmp_def->refBy.end();) {

	  def_refBy = *it_ref;
	  //print_ad_ptuf(def_refBy);
	  elim_prop_tr_un_fl(def_refBy->rhs_tr_fl, tmp_def->rhs_tr_fl, aiger_lit2var(tmp_def->lhs));

	  update_refby_un_hl_ve(def_refBy->rhs_tr_fl, tmp_def, def_refBy);

	  def_refBy->nb_op += tmp_def->nb_op - 1;

	  it_ref = tmp_def->refBy.erase(it_ref);

	}
	assert(!tmp_def->refBy.size());

	//dll_remove_elem(list_ands[time], tmp_e);
	dll_remove_elem(tmp_e);

	stats_un_hl_ve_cl++;
      }
    }
    tmp_e = tmp_e->next;
  }
}

int add_ref(prop_tr_un_fl *ptuf, ands_def *ad) {
  int ret = 0;
  if (ptuf->op == VAR) {
    if (ptuf->lit_ref != NULL) {
      ptuf->lit_ref->refBy.insert(ad);
    }
    ret++;
  } else {
    list_ptuf_e *it = ptuf->lp->head;
    while(it != NULL) {
      ret += add_ref(it->ptuf, ad);
      it = it->next;
    }
  }
  return ret;
}
void build_refBy_ad(int time) {
  ands_def* tmp_def;
  dll_elem<ands_def> *tmp_e;
  tmp_e = list_ands[time]->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    tmp_def->refBy.clear();
    tmp_def->nb_op = add_ref(tmp_def->rhs_tr_fl, tmp_def);

    tmp_e = tmp_e->next;
  }
}

void un_hl_ve(bool verbose, int time) {
  clock_t start = clock();

  set_ad_apply_ve(time);

  build_refBy_ad(time);

  un_hl_ve_ClausesBased(time);

  reset_ad_apply_ve(time);
  
  if (verbose) printf("[MC]  Unrolled High level VE : %d (%.2fs)\n",
		      stats_un_hl_ve_cl, ((double) (clock() - start)) / CLOCKS_PER_SEC);
}


void un_hl_ve_zz(bool verbose, int time) {
  clock_t start = clock();

  set_ad_apply_ve(time);

  build_refBy_ad(time);

  un_hl_ve_ClausesBased(time);
  
  if (verbose) printf("[MC]  Unrolled High level VE : %d (%.2fs)\n",
		      stats_un_hl_ve_cl, ((double) (clock() - start)) / CLOCKS_PER_SEC);
}
